"""
The module ``record`` provides the equivalent of a logger for array data.

===============================================================================
Overview
===============================================================================
Applying the decorator ``static_subgrouping_array_recorders`` with the names of
desired recorders as strings or keyword arguments with a given recorder
instance initializes the ``recorders`` attribute, which contains each recorder
as a dictionary would. Results are stored under a group named after decorated
function. Nested function calls can be paralleled by nesting results within
groups. All that is required is that the recorder be assigned from outer scope
to inner (see example below).

.. code:: python

    @static_subgrouping_array_recorders("r")
    def a(x):
        a.r = x
        return x * x.T

    @static_subgrouping_array_recorders("r")
    def b(x, y):
        a.r = b.r
        return a(x + y)

All recorders default to an instance of ``EmptyArrayRecorder`` unless otherwise
specified. All operations work on them exactly the same, but no recording
occurs. This guarantees that code using recorders will not break if they don't
have them. A similar strategy can be applied for classes and their methods by
using ``class_static_array_debug_recorder``.

.. todo:: Complete documentation with more examples.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jun 04, 2014 11:10:55 EDT$"


import numpy
import h5py

from nanshe.io.hdf5 import serializers
from nanshe.util import wrappers


# Need in order to have logging information no matter what.
from nanshe.util import prof


# Get the logger
trace_meta_logger = prof.getTraceMetaLogger(__name__)


@prof.log_class(trace_meta_logger)
class EmptyArrayRecorder(object):
    def __init__(self):
        self.__recorders = set()

    def __nonzero__(self):
        return(False)

    # For forward compatibility with Python 3
    __bool__ = __nonzero__

    def get(self, key, default=None):
        value = default

        try:
            value = self.__getitem__(key)
        except KeyError:
            pass

        return(value)

    def __contains__(self, key):
        return(key in self.__recorders)

    def __getitem__(self, key):
        if (key == "."):
            return(self)

        if key in self.__recorders:
            return(EmptyArrayRecorder())
        else:
            raise(KeyError(
                "unable to open object (Symbol table: Can't open object " +
                repr(key) + ")"
            ))

    def __setitem__(self, key, value):
        # Exception will be thrown if value is empty or if key already exists
        # (as intended).
        if (key == "."):
            if not ((value is None) or (value is h5py.Group)):
                raise ValueError("Cannot store dataset in top level group.")
        elif (value is None) or (value is h5py.Group):
            self.__recorders.add(key)
        else:
            if value.size:
                pass
            else:
                raise ValueError(
                    "The array provided for output by the name: \"" + key +
                    "\" is empty."
                )


@prof.log_class(trace_meta_logger)
class HDF5ArrayRecorder(object):
    def __init__(self, hdf5_handle, overwrite=False):
        self.hdf5_handle = hdf5_handle
        self.overwrite = overwrite

    def __nonzero__(self):
        return(True)

    # For forward compatibility with Python 3
    __bool__ = __nonzero__

    def get(self, key, default=None):
        value = default

        try:
            value = self.__getitem__(key)
        except KeyError:
            pass

        return(value)

    def __contains__(self, key):
        return(key in self.hdf5_handle)

    def __getitem__(self, key):
        if (key == "."):
            return(self)

        try:
            if isinstance(self.hdf5_handle[key], h5py.Group):
                return(HDF5ArrayRecorder(
                    self.hdf5_handle[key], overwrite=self.overwrite
                ))
            else:
                return(serializers.read_numpy_structured_array_from_HDF5(
                    self.hdf5_handle, key
                ))
        except:
            raise(KeyError(
                "unable to open object (Symbol table: Can't open object " +
                repr(key) + " in " + repr(self.hdf5_handle) + ")"
            ))

    def __setitem__(self, key, value):
        if (key == "."):
            if not ((value is None) or (value is h5py.Group)):
                raise ValueError(
                    "Cannot store dataset in top level group ( " +
                    self.hdf5_handle.name + " )."
                )

            if self.overwrite:
                for each_key in self.hdf5_handle:
                    del self.hdf5_handle[each_key]

                self.hdf5_handle.file.flush()
        elif (value is None) or (value is h5py.Group):
            # Check to see if the output must go somewhere special.
            if key:
                # If so, check to see if it exists.
                if key in self.hdf5_handle:
                    # If it does and we want to overwrite it, do so.
                    if self.overwrite:
                        del self.hdf5_handle[key]

                        self.hdf5_handle.create_group(key)

                        self.hdf5_handle.file.flush()
                else:
                    # Create it if it doesn't, exist.
                    self.hdf5_handle.create_group(key)

                    self.hdf5_handle.file.flush()
        else:
            # Attempt to create a dataset in self.hdf5_handle named key with
            # value and do not overwrite. Exception will be thrown if value is
            # empty or if key already exists (as intended).
            if value.size:
                serializers.create_numpy_structured_array_in_HDF5(
                    self.hdf5_handle,
                    key,
                    value,
                    overwrite=self.overwrite
                )
                self.hdf5_handle.file.flush()

                return()
            else:
                raise ValueError(
                    "The array provided for output by the name: \"" +
                    key + "\" is empty."
                )


@prof.log_class(trace_meta_logger)
class HDF5EnumeratedArrayRecorder(object):
    def __init__(self, hdf5_handle):
        self.hdf5_handle = hdf5_handle

        # Must be a logger if it already exists.
        assert self.hdf5_handle.attrs.get("is_logger", True)

        self.hdf5_handle.attrs["is_logger"] = True
        self.hdf5_handle.file.flush()

        self.hdf5_index_data_handles = {"." : -1}
        for each_index in self.hdf5_handle:
            each_index = int(each_index)
            self.hdf5_index_data_handles["."] = max(
                self.hdf5_index_data_handles["."], each_index
            )

        if self.hdf5_index_data_handles["."] != -1:
            hdf5_index_handle = self.hdf5_handle[str(
                self.hdf5_index_data_handles["."]
            )]
            for each_key in hdf5_index_handle:
                if hdf5_index_handle[each_key].attrs.get("is_logger", False):
                    self.hdf5_index_data_handles[each_key] = None
                else:
                    self.hdf5_index_data_handles[each_key] = -1

                    for each_index in hdf5_index_handle[each_key]:
                        each_index = int(each_index)
                        self.hdf5_index_data_handles[each_key] = max(
                            self.hdf5_index_data_handles[each_key], each_index
                        )

    def __nonzero__(self):
        return(True)

    # For forward compatibility with Python 3
    __bool__ = __nonzero__

    def get(self, key, default=None):
        value = default

        try:
            value = self.__getitem__(key)
        except KeyError:
            pass

        return(value)

    def __contains__(self, key):
        return(self.get(key) is not None)

    def __getitem__(self, key):
        if (key == "."):
            return(self)
        try:
            root_i = self.hdf5_index_data_handles.get(".", -1)
            key_i = self.hdf5_index_data_handles.get(key, -1)

            assert (root_i != -1)
            assert (key_i != -1)

            root_i_str = str(root_i)

            assert isinstance(self.hdf5_handle[root_i_str], h5py.Group)
            assert isinstance(self.hdf5_handle[root_i_str][key], h5py.Group)

            key_handle = self.hdf5_handle[root_i_str][key]
            if key_i is None:
                return(HDF5EnumeratedArrayRecorder(key_handle))
            else:
                key_i_str = str(key_i)
                return(serializers.read_numpy_structured_array_from_HDF5(
                    key_handle, key_i_str
                ))
        except:
            raise(KeyError(
                "unable to open object (Symbol table: Can't open object " +
                repr(key) + " in " + repr(self.hdf5_handle) + ")"
            ))

    def __setitem__(self, key, value):
        if (key == "."):
            if not ((value is None) or (value is h5py.Group)):
                raise ValueError(
                    "Cannot store dataset in top level group ( " +
                    self.hdf5_handle.name + " )."
                )

            self.hdf5_index_data_handles = {
                "." : self.hdf5_index_data_handles["."] + 1
            }
            self.hdf5_handle.create_group(
                str(self.hdf5_index_data_handles["."])
            )
            self.hdf5_handle.file.flush()
        else:
            hdf5_index_handle = None
            try:
                hdf5_index_handle = self.hdf5_handle[str(
                    self.hdf5_index_data_handles["."]
                )]
            except KeyError:
                if self.hdf5_index_data_handles["."] == -1:
                    self.hdf5_index_data_handles = {
                        "." : self.hdf5_index_data_handles["."] + 1
                    }

                self.hdf5_handle.create_group(
                    str(self.hdf5_index_data_handles["."])
                )
                self.hdf5_handle.file.flush()

                hdf5_index_handle = self.hdf5_handle[str(
                    self.hdf5_index_data_handles["."]
                )]

            if (value is None) or (value is h5py.Group):
                # Create a group if it doesn't already exist.
                hdf5_index_handle.require_group(key)
                hdf5_index_handle.attrs["is_logger"] = True
                hdf5_index_handle.file.flush()
                self.hdf5_index_data_handles[key] = None
            else:
                # Index into a NumPy structured array can return a void type
                # even though it is a valid array, which can be stored.
                # So, we must check.
                try:
                    assert isinstance(value, numpy.ndarray)
                except AssertionError:
                    if not value.dtype.names:
                        raise
                if value.size:
                    # If so, check to see if it exists.
                    if key not in self.hdf5_index_data_handles:
                        hdf5_index_handle.create_group(key)
                        hdf5_index_handle[key].attrs["is_logger"] = False
                        hdf5_index_handle.file.flush()
                        self.hdf5_index_data_handles[key] = -1

                    self.hdf5_index_data_handles[key] += 1

                    serializers.create_numpy_structured_array_in_HDF5(
                        hdf5_index_handle[key],
                        str(self.hdf5_index_data_handles[key]),
                        value
                    )

                    self.hdf5_handle.file.flush()
                else:
                    raise ValueError(
                        "The array provided for output by the name: \"" +
                        key + "\" is empty."
                    )


@prof.log_call(trace_meta_logger)
def generate_HDF5_array_recorder(hdf5_handle,
                                 group_name="",
                                 enable=True,
                                 overwrite_group=False,
                                 recorder_constructor=HDF5ArrayRecorder,
                                 **kwargs):
    """
        Generates a function used for writing arrays (structured or otherwise)
        to a group in an HDF5 file.

        Args:
            hdf5_handle:            The HDF5 file group to place the debug
                                    contents into.

            group_name:             The name of the group within hdf5_handle to
                                    save the contents to. (If set to the empty
                                    string, data will be saved to
                                    hdf5_handle directly)

            enable:                 Whether to generate a real recorder or a
                                    fake one.

            overwrite_group:        Whether to overwrite the group where data
                                    is stored.

            recorder_constructor:   Type of recorder to use if enable is True.

            **kwargs:               Other arguments to pass through to the
                                    recorder_constructor (won't pass through if
                                    enable is false).

        Returns:
            ArrayRecorder:          A function, which will take a given array
                                    name and value and write them out.
    """

    if isinstance(hdf5_handle, str):
        hdf5_handle = h5py.File(hdf5_handle, "a")

    if (enable):
        hdf5_recording_handle = hdf5_handle

        # Check to if the output must go somewhere special.
        if group_name:
            # If so, check to see if it exists.
            if group_name in hdf5_handle:
                # If it does and we want to overwrite it, do so.
                if overwrite_group:
                    del hdf5_handle[group_name]

                    hdf5_handle.create_group(group_name)

                    hdf5_handle.file.flush()
            else:
                # Create it if it doesn't, exist.
                hdf5_handle.create_group(group_name)

                hdf5_handle.file.flush()

            hdf5_recording_handle = hdf5_handle[group_name]

        return(recorder_constructor(hdf5_recording_handle, **kwargs))
    else:
        return(EmptyArrayRecorder())


@prof.log_call(trace_meta_logger)
def static_subgrouping_array_recorders(*args, **kwargs):
    """
        Creates a decorator that adds a static variable, recorders, that holds
        as many recorders as are supplied.

        Args:
            args(tuple of strs):                        All variables to be
                                                        named (set to
                                                        EmptyArrayRecorder()).

        Keyword Args:
            kwargs(dict of strs and ArrayRecorders):    All variables to be
                                                        named with values of
                                                        type ArrayRecorder.

        Returns:
            (callable):                                 A decorator that adds
                                                        the static variable,
                                                        recorders, to the given
                                                        function.
    """

    @prof.log_call(trace_meta_logger)
    def static_subgrouping_array_recorders_tie(callable):
        """
            Creates a decorator that adds a static variable recorders to the
            function it decorates.

            Args:
                callable(callable):     All variables to be named (set to
                                        EmptyArrayRecorder()).

            Returns:
                (callable):             A function with the static variable,
                                        recorders, added.
        """

        class SubgroupingRecorders(object):
            # """
            #     Hold recorders. Automatically, moves instances of
            #     ArrayRecorder to a subgroup with the same name as the
            #     callable on assignment.
            # """
            def __init__(self, *args, **kwargs):
                # """
                #     Contains ArrayRecorders that write to a subgroup of the
                #     same name as the callable.
                #
                #     Args:
                #         args(tuple of strs):                        All
                #                                                     variables
                #                                                     to be
                #                                                     named
                #                                                     (set to
                #                                                     EmptyArrayRecorder()).
                #
                #     Keyword Args:
                #         kwargs(dict of strs and ArrayRecorders):    All
                #                                                     variables
                #                                                     to be
                #                                                     named
                #                                                     with
                #                                                     values of
                #                                                     type
                #                                                     ArrayRecorder.
                # """

                for _k in args:
                    object.__setattr__(self, _k, None)

                for _k, _v in kwargs.items():
                    object.__setattr__(self, _k, _v)

            def __getattr__(self, _k):
                if _k != "__dict__":
                    return(self.__dict__[_k])

            def __setattr__(self, _k, _v):
                if _k != "__dict__":
                    if _v is None:
                        self.__dict__[_k] = EmptyArrayRecorder()
                    else:
                        _v[callable.__name__] = None
                        try:
                            self.__dict__[_k] = _v[callable.__name__]
                        except KeyError:
                            if isinstance(_v, EmptyArrayRecorder):
                                self.__dict__[_k] = EmptyArrayRecorder()
                            else:
                                raise

            def __delattr__(self, _k):
                if _k != "__dict__":
                    del self.__dict__[_k]

        callable = wrappers.static_variables(
            recorders=SubgroupingRecorders(*args, **kwargs)
        )(callable)

        @wrappers.wraps(callable)
        def static_subgrouping_array_recorders_wrapper(*args, **kwargs):
            # Force all recorders to ensure their output Group exists.
            # All of them actually make the directory.
            # However, HDF5EnumeratedArrayRecorder needs a clue as to
            # when it should switch to a new one as it will keep different
            # runs separate.
            for _k in callable.recorders.__dict__:
                callable.recorders.__dict__[_k]["."] = None

            return(callable(*args, **kwargs))

        return(static_subgrouping_array_recorders_wrapper)

    return(static_subgrouping_array_recorders_tie)


@prof.log_call(trace_meta_logger)
def static_array_debug_recorder(callable):
    """
        Creates a decorator that adds a static variable recorders that contains
        the variable array_debug_recorder to the function it decorates.
        By
        default, array_debug_recorder is set to an EmptyArrayRecorder instance.
        Also, on assignment it automatically creates a subgroup with the same
        name as the function.

        Args:
            callable(callable):    All variables to be named (set to
                                   EmptyArrayRecorder()).

        Returns:
            (callable):            A decorator that adds the static variable
                                   array_debug_recorder to the given function.
    """

    callable = static_subgrouping_array_recorders(
        array_debug_recorder=EmptyArrayRecorder()
    )(callable)

    return(callable)


@prof.log_call(trace_meta_logger)
def class_static_subgrouping_array_recorders(*args, **kwargs):
    """
        Creates a decorator that adds a static variable, recorders, that holds
        as many recorders as are supplied.

        Args:
            args(tuple of strs):                        All variables to be
                                                        named (set to
                                                        EmptyArrayRecorder()).

        Keyword Args:
            kwargs(dict of strs and ArrayRecorders):    All variables to be
                                                        named with values of
                                                        type ArrayRecorder.

        Returns:
            (callable):                                 A decorator that adds
                                                        the static variable,
                                                        recorders, to the given
                                                        function.
    """

    @prof.log_call(trace_meta_logger)
    def class_static_subgrouping_array_recorders_tie(a_class):
        """
            Creates a decorator that adds a static variable recorders to the
            function it decorates.

            Args:
                a_class(class):      All variables to be named (set to
                                     EmptyArrayRecorder()).

            Returns:
                (class):             A function with the static variable,
                                     recorders, added.
        """

        class ClassSubgroupingRecorders(object):
            # """
            #     Hold recorders. Automatically, moves instances of
            #     ArrayRecorder to a subgroup with the same name as the a_class
            #     on assignment.
            # """
            def __init__(self, *args, **kwargs):
                # """
                #     Contains ArrayRecorders that write to a subgroup of the
                #     same name as the a_class.
                #
                #     Args:
                #         args(tuple of strs):                       All
                #                                                    variables
                #                                                    to be named
                #                                                    (set to
                #                                                    EmptyArrayRecorder()).
                #
                #     Keyword Args:
                #         kwargs(dict of strs and ArrayRecorders):   All
                #                                                    variables
                #                                                    to be named
                #                                                    with values
                #                                                    of type
                #                                                    ArrayRecorder.
                # """

                for _k in args:
                    object.__setattr__(self, _k, None)

                for _k, _v in kwargs.items():
                    object.__setattr__(self, _k, _v)

            def __getattr__(self, _k):
                if _k != "__dict__":
                    return(self.__dict__[_k])

            def __setattr__(self, _k, _v):
                if _k != "__dict__":
                    if _v is None:
                        self.__dict__[_k] = EmptyArrayRecorder()
                    else:
                        _v[a_class.__name__] = None
                        try:
                            self.__dict__[_k] = _v[a_class.__name__]
                        except KeyError:
                            if isinstance(_v, EmptyArrayRecorder):
                                self.__dict__[_k] = EmptyArrayRecorder()
                            else:
                                raise

            def __delattr__(self, _k):
                if _k != "__dict__":
                    del self.__dict__[_k]

        a_class = wrappers.class_static_variables(
            recorders=ClassSubgroupingRecorders(*args, **kwargs)
        )(a_class)

        def class_static_subgrouping_array_recorders__init__decorator(callable):
            @wrappers.wraps(callable)
            def class_static_subgrouping_array_recorders__init__wrapper(self,
                                                                        *args,
                                                                        **kwargs):
                # Force all recorders to ensure their output Group exists.
                # All of them actually make the directory.
                # However, HDF5EnumeratedArrayRecorder needs a clue as to
                # when it should switch to a new one as it will keep different
                # runs separate.

                self.recorders = ClassSubgroupingRecorders()
                for _k in a_class.recorders.__dict__:
                    self.recorders.__dict__[_k] = a_class.recorders.__dict__[_k]

                return(callable(self, *args, **kwargs))

            return(class_static_subgrouping_array_recorders__init__wrapper)

        def class_static_subgrouping_array_recorders_decorator(callable):
            callable = static_subgrouping_array_recorders()(callable)

            @wrappers.wraps(callable)
            def class_static_subgrouping_array_recorders_wrapper(self,
                                                                 *args,
                                                                 **kwargs):
                # Force all recorders to ensure their output Group exists.
                # All of them actually make the directory.
                # However, HDF5EnumeratedArrayRecorder needs a clue as to
                # when it should switch to a new one as it will keep different
                # runs separate.
                callable.recorders.__dict__ = dict()
                for _k in self.recorders.__dict__:
                    # setattr(callable.recorders, _k,
                    # self.recorders.__dict__[_k][callable.__name__)])
                    setattr(
                        callable.recorders, _k, getattr(self.recorders, _k)
                    )
                    # callable.recorders.__dict__[_k]["."] = None

                return(callable(self, *args, **kwargs))

            return(class_static_subgrouping_array_recorders_wrapper)

        # Wraps __init__ only
        a_class = wrappers.class_decorate_methods(
            __init__=class_static_subgrouping_array_recorders__init__decorator
        )(a_class)

        # Wrap everything
        a_class = wrappers.class_decorate_all_methods(
            class_static_subgrouping_array_recorders_decorator)(a_class)


        # Must be done last.
        # Precedes the constructor to ensure a new working directory is created
        class MetaSubgroupingRecorders(type):
            def __call__(self, *args, **kwargs):
                for _k in self.recorders.__dict__:
                    self.recorders.__dict__[_k]["."] = None

                return(super(MetaSubgroupingRecorders, self).__call__(
                    *args, **kwargs
                ))

        a_class = wrappers.metaclass(MetaSubgroupingRecorders)(a_class)

        return(a_class)

    return(class_static_subgrouping_array_recorders_tie)


@prof.log_call(trace_meta_logger)
def class_static_array_debug_recorder(a_class):
    """
        Creates a decorator that adds a static variable recorders that contains
        the variable array_debug_recorder to the function it decorates. By
        default, array_debug_recorder is set to an EmptyArrayRecorder instance.
        Also, on assignment it automatically creates a subgroup with the same
        name as the function.

        Args:
            a_class(class):    All variables to be named
                               (set to EmptyArrayRecorder()).

        Returns:
            (class):           A decorator that adds the static variable
                               array_debug_recorder to the given function.
    """

    a_class = class_static_subgrouping_array_recorders(
        array_debug_recorder=EmptyArrayRecorder()
    )(a_class)

    return(a_class)



"""
The module ``xglob`` extends the abilities found in Python's ``glob``.

===============================================================================
Overview
===============================================================================
The module ``xglob`` extends the abilities found in |glob|_. In particular, it
provides a function for iterating through several glob expressions and joining
them together.

.. |glob| replace:: ``glob``
.. _glob: http://docs.python.org/2/library/glob.html

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jun 26, 2014 18:24:25 EDT$"


import glob

from nanshe.util import prof



# Get the logger
trace_logger = prof.getTraceLogger(__name__)



@prof.log_call(trace_logger)
def expand_pathname_list(*pathnames):
    """
        Takes each pathname in those given and expands them using regex.

        Args:
            *pathnames(str):     pathnames to use regex on to expand.

        Returns:
            list:                a list of path names (without regex)
    """

    expanded_pathnames = []

    # Completes any regex
    expanded_pathnames = []
    for each_pathname in pathnames:
        expanded_pathnames.extend(glob.glob(each_pathname))

    return(expanded_pathnames)



"""
The ``neuron_matplotlib_viewer`` module provides a ``matplotlib``-based viewer.

===============================================================================
Overview
===============================================================================
The module ``neuron_matplotlib_viewer`` provides a simple |matplotlib|_ viewer
for navigating through a 3D imagestack (TYX). However the first dimension could
be Z, as well. This has been deprecated in favor of the ``viewer``.

.. |matplotlib| replace:: ``matplotlib``
.. _matplotlib: http://matplotlib.org

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Apr 03, 2014 20:20:39 EDT$"


import warnings

warnings.warn(
    "The module `neuron_matplotlib_viewer` is deprecated."
    "Please consider using `viewer` instead.",
    DeprecationWarning
)

# Need in order to have logging information no matter what.
from nanshe.util import prof

# To get type info.
from nanshe.util.xnumpy import info

# Get the logger
trace_logger = prof.getTraceLogger(__name__)
logger = prof.logging.getLogger(__name__)

import matplotlib as mpl
import matplotlib.figure
import matplotlib.colors

from matplotlib.widgets import Slider, Button, SpanSelector


@prof.log_class(trace_logger)
class NeuronMatplotlibViewer(matplotlib.figure.Figure):
    """
        Provides a way to interact with numpy arrays pulled from neuron images.

        Wraps a Matplotlib figure instance.
    """

    def __init__(self, *args, **kwargs):
        """
            Initializes a NeuronMatplotlibViewer using the given figure to
            clone and image stack to view.

            After cloning, self will be the same as fig. Additional features
            will be attached to self.

            Args:
                args(list):         arguments to be passed to parent

            Keyword Args:
                kwargs(dict):       arguments to be passed to parent
        """

        super(NeuronMatplotlibViewer, self).__init__(*args, **kwargs)
        super(NeuronMatplotlibViewer, self).__init__(
            *args, **dict([(_k, _v) for (_k, _v) in kwargs.items() if _k is not "neuron_images"])
        )

        self.subplots_adjust(left=0.25, bottom=0.25)
        self.viewer = self.add_axes([0.25, 0.25, 0.7, 0.7])

        #self.set_images(kwargs["neuron_images"])

    def set_images(self,
                   new_neuron_images,
                   cmap=mpl.cm.Greys_r,
                   use_matshow=False,
                   vmin=None,
                   vmax=None):
        """
            Sets the images to be viewed.

            Args:
                new_neuron_images(numpy.ndarray):     array of images (first
                                                      index is which image)
        """
        if (len(new_neuron_images.shape) > 3):
            raise ValueError(
                "Dimensions cannot be greater than 3. " +
                "Was provided new_neuron_images with \"" +
                str(len(new_neuron_images.shape)) + "\" dimensions."
            )

        self.neuron_images = new_neuron_images
        self.cmap = cmap
        self.vmin = vmin
        self.vmax = vmax

        type_info = info(new_neuron_images.dtype)
        if self.vmin is None:
           self.vmin = type_info.min
        if self.vmax is None:
           self.vmax = type_info.max

        self.svmin = self.vmin
        self.svmax = self.vmax

        viewer_show_method = None
        if use_matshow:
            viewer_show_method = self.viewer.matshow
        else:
            viewer_show_method = self.viewer.imshow

        self.image_view = viewer_show_method(
            self.get_image(0),
            cmap=self.cmap,
            vmin=self.vmin,
            vmax=self.vmax
        )

        self.image_view_colorbar = self.colorbar(
            self.image_view, ax=self.viewer)

        self.image_view_colorbar_selector = SpanSelector(
            self.image_view_colorbar.ax,
            self.color_range_update, 'vertical',
            useblit=True,
            rectprops=dict(alpha=0.3, facecolor='yellow')
        )

        if (len(self.neuron_images.shape) == 3):
            self.time_nav = TimeNavigator(self, len(self.neuron_images) - 1)

            self.time_nav_cid = self.time_nav.on_time_update(self.time_update)

        self.viewer.format_coord = self.format_coord

    def get_image(self, i=None):
        """
            Gets the current image or the image if it is a projection.

            Args:
                i(int):             image to retrieve (defaults to selection).

            Returns:
                numpy.ndarray:      the current image.
        """

        cur_img = self.neuron_images
        if (len(self.neuron_images.shape) == 3):
            i = self.time_nav.stime.val if i is None else i
            cur_img = cur_img[i]
        else:
            cur_img = cur_img[...]

        cur_img = cur_img[...]

        cur_img = cur_img.astype(float)

        return(cur_img)

    def color_range_update(self, vmin, vmax):
        """
            Handles an update to the vmin and vmax range based on the selection
            provided.

            Args:
                vmin         the min value selected
                vmax         the max value selected
        """
        if vmin != vmax:
            self.svmin = (self.svmax - self.svmin) * vmin + self.svmin
            self.svmax = (self.svmax - self.svmin) * vmax + self.svmin
        else:
            self.svmin = self.vmin
            self.svmax = self.vmax

        norm = matplotlib.colors.Normalize(self.svmin, self.svmax)
        self.image_view_colorbar.set_norm(norm)
        self.image_view.set_norm(norm)

        self.canvas.draw_idle()

    def time_update(self):
        """
            Method to be called by the TimeNavigator when the time changes.
            Updates image displayed.
        """
        if (len(self.neuron_images.shape) == 3):
            self.image_view.set_array(self.get_image())
            self.canvas.draw_idle()

    def format_coord(self, x, y):
        """
            Include intensity when showing coordinates during mouseover.

            Args:
                x(float):    cursor's x position within the image.
                y(float):    cursor's y position within the image.

            Returns:
                str:         coordinates and intensity if it can be gotten.
        """
        try:
            xi = int(round(x))
            yi = int(round(y))

            z = self.get_image()[yi, xi]

            return 'x=%1.4f, y=%1.4f, z=%1.4f'%(x, y, z)
        except:
            return 'x=%1.4f, y=%1.4f'%(x, y)


@prof.log_class(trace_logger)
class TimeNavigator(object):
    def __init__(self, fig, max_time, min_time=0, time_step=1, axcolor='lightgoldenrodyellow',
                 hovercolor='0.975'):
        """
            Initializes a TimeNavigator using the given figure and a fixed
            number of steps.

            Provides a simple slider bar and buttons to navigate through time.

            Args:
                fig         should be a figure that has been initialized.
                max_time    maximum step for images.
                min_time    minimum step for images.
                time_step   how much to increase/decrease by for each step.
                axcolor     color to be used for all buttons and slider bar.
                hovercolor  color turned when mouse over occurs for any button.
        """

        self.min_time = min_time
        self.max_time = max_time
        self.time_step = time_step
        self.axcolor = axcolor
        self.hovercolor = hovercolor

        self.next_cid = 0

        self.axtime = fig.add_axes(
            [0.25, 0.1, 0.65, 0.03], axisbg=self.axcolor)
        self.stime = Slider(
            self.axtime,
            'Time',
            self.min_time,
            self.max_time,
            valinit=self.min_time,
            valfmt='%i'
        )

        self.stime.on_changed(self.time_update)

        self.beginax = fig.add_axes([0.2, 0.025, 0.1, 0.04])
        self.begin_button = Button(
            self.beginax,
            'Begin',
            color=self.axcolor,
            hovercolor=self.hovercolor
        )
        self.begin_button.on_clicked(self.begin_time)

        self.prevax = fig.add_axes([0.3, 0.025, 0.1, 0.04])
        self.prev_button = Button(
            self.prevax, 'Prev', color=self.axcolor, hovercolor=self.hovercolor
        )
        self.prev_button.on_clicked(self.prev_time)

        self.nextax = fig.add_axes([0.7, 0.025, 0.1, 0.04])
        self.next_button = Button(
            self.nextax, 'Next', color=self.axcolor, hovercolor=self.hovercolor
        )
        self.next_button.on_clicked(self.next_time)

        self.endax = fig.add_axes([0.8, 0.025, 0.1, 0.04])
        self.end_button = Button(
            self.endax, 'End', color=self.axcolor, hovercolor=self.hovercolor
        )
        self.end_button.on_clicked(self.end_time)

        self.callbacks = {}

    def begin_time(self, event):
        """
            Sets time to min_time.

            Args:
                event   Matplotlib event that caused the call to this callback.
        """

        logger.debug(
            "Value of slider before setting to the beginning is \"" +
            str(self.stime.val) + "\"."
        )

        self.time_update(self.min_time)

        logger.debug(
            "Value of slider after setting to the beginning is \"" +
            str(self.stime.val) + "\"."
        )
        assert (self.min_time == self.stime.val)

    def prev_time(self, event):
        """
            Sets time to one time_step prior.

            Args:
                event   Matplotlib event that caused the call to this callback.
        """

        logger.debug(
            "Value of slider before going to the previous time is \"" +
            str(self.stime.val) + "\"."
        )

        self.time_update(self.stime.val - self.time_step)

        logger.debug(
            "Value of slider after going to the previous time is \"" +
            str(self.stime.val) + "\"."
        )

    def next_time(self, event):
        """
            Sets time to one time_step after.

            Args:
                event   Matplotlib event that caused the call to this callback.
        """

        logger.debug(
            "Value of slider before going to the next time is \"" +
            str(self.stime.val) + "\"."
        )

        self.time_update(self.stime.val + self.time_step)

        logger.debug(
            "Value of slider after going to the next time is \"" +
            str(self.stime.val) + "\"."
        )

    def end_time(self, event):
        """
            Sets time to max_time.

            Args:
                event   Matplotlib event that caused the call to this callback.
        """

        self.time_update(self.max_time)

        assert (self.max_time == self.stime.val)

    def normalize_val(self, val):
        """
            Takes the time value and normalizes it to fit within the range.
            Then, makes sure it is a discrete number of steps from the
            min_time.

            Args:
                val     float position from the slider bar to correct

            Returns:
                int:    the normalized value.
        """

        if val < self.min_time:
            return(self.min_time)
        elif val > self.max_time:
            return(self.max_time)
        else:
            return(int(round((val - self.min_time) / self.time_step)))

    def time_update(self, val):
        """
            Takes the time value and normalizes it within the range if it does
            not fit.

            Args:
                val     float position from slider bar to move to
        """

        val = self.normalize_val(val)

        if val != self.stime.val:
            self.stime.set_val(val)

            for each_cid, each_callback in self.callbacks.items():
                logger.debug(
                    "Before calling the caller id for time_update with value \"" +
                    str(each_cid) + "\"."
                )

                each_callback()

                logger.debug(
                    "After calling the caller id for time_update with value \"" +
                    str(each_cid) + "\"."
                )

    def disconnect(self, cid):
        """
            Disconnects the given cid from being notified of time updates.

            Args:
                cid     ID of callback to pull
        """

        logger.debug(
            "Before disconnecting the caller id for time_update with value \"" +
            str(cid) + "\"."
        )
        logger.debug(
            "Contents of the callback dictionary before disconnecting \"" +
            str(self.callbacks) + "\"."
        )

        del self.callbacks[cid]

        logger.debug(
            "After disconnecting the caller id for time_update with value \"" +
            str(cid) + "\"."
        )
        logger.debug(
            "Contents of the callback dictionary after disconnecting\"" +
            str(self.callbacks) + "\"."
        )

    def on_time_update(self, func):
        """
            Registers a callback function for notification when the time is
            updated.

            Args:
                func(callable):     function call when the time is updated

            Returns:
                int:                a callback ID or cid to allow pulling the
                                    callback when no longer necessary.
        """

        logger.debug(
            "Before connecting the caller id for time_update with value \"" +
            str(self.next_cid) + "\"."
        )
        logger.debug(
            "Contents of the callback dictionary before connecting \"" +
            str(self.callbacks) + "\"."
        )

        cid = self.next_cid
        self.next_cid += 1

        self.callbacks[cid] = func

        logger.debug(
            "After connecting the caller id for time_update with value \"" +
            str(cid) + "\"."
        )
        logger.debug(
            "Contents of the callback dictionary after connecting \"" +
            str(self.callbacks) + "\"."
        )

        return(cid)



"""
The package ``misc`` contains everything that does not really belong elsewhere.

===============================================================================
Overview
===============================================================================
The package ``misc`` (short for **misc**\ ellaneous) contains random components
that did not really belong anywhere else. This stuff is likely not well
documented or tested. As a consequence, it is safe to assume that this stuff
is not stable and may be deprecated. If does ever leave this package (and is
not deleted), its API may be broken from what it was here.
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Apr 01, 2015 00:13:16 EDT$"

__all__ = [  # "neuron_matplotlib_viewer",
    "random_dictionary_learning_data"
]

# from nanshe.misc import neuron_matplotlib_viewer
from nanshe.misc import random_dictionary_learning_data



"""
The ``registration`` module helps bring image stacks into alignment.

===============================================================================
Overview
===============================================================================
The registration algorithm developed by Wenzhi Sun. The strategy uses an
area-based method for registration. Namely, it takes a template image (the mean
projection) and finds the translations required for each frame to overlap
optimally with the template image. Then a number of adjustments are made to the
shifts to ensure that unnecessary translations are not preformed.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jan 28, 2015 11:25:47 EST$"


import itertools
import os
import tempfile
import warnings

import h5py
import numpy


try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache

try:
    import pyfftw.interfaces.numpy_fft as fft
except Exception as e:
    warnings.warn(str(e) + ". Falling back to NumPy FFTPACK.", ImportWarning)
    import numpy.fft as fft

from nanshe.util import iters, xnumpy
from nanshe.io import hdf5

# Need in order to have logging information no matter what.
from nanshe.util import prof


# Get the loggers
trace_logger = prof.getTraceLogger(__name__)
logger = prof.logging.getLogger(__name__)



@prof.log_call(trace_logger)
def register_mean_offsets(frames2reg,
                          max_iters=-1,
                          block_frame_length=-1,
                          include_shift=False,
                          to_truncate=False,
                          float_type=numpy.dtype(float).type):
    """
        This algorithm registers the given image stack against its mean
        projection. This is done by computing translations needed to put each
        frame in alignment. Then the translation is performed and new
        translations are computed. This is repeated until no further
        improvement can be made.

        The code for translations can be found in find_mean_offsets.

        Notes:
            Adapted from code provided by Wenzhi Sun with speed improvements
            provided by Uri Dubin.

        Args:
            frames2reg(numpy.ndarray):           Image stack to register (time
                                                 is the first dimension uses
                                                 C-order tyx or tzyx).

            max_iters(int):                      Number of iterations to allow
                                                 before forcing termination if
                                                 stable point is not found yet.
                                                 Set to -1 if no limit.
                                                 (Default -1)

            block_frame_length(int):             Number of frames to work with
                                                 at a time. By default all.
                                                 (Default -1)

            include_shift(bool):                 Whether to return the shifts
                                                 used, as well. (Default False)

            to_truncate(bool):                   Whether to truncate the frames
                                                 to remove all masked portions.
                                                 (Default False)

            float_type(type):                    Type of float to use for
                                                 calculation. (Default
                                                 numpy.float64).

        Returns:
            (numpy.ndarray):                     an array containing the
                                                 translations to apply to each
                                                 frame.

        Examples:
            >>> a = numpy.zeros((5, 3, 4)); a[:,0] = 1; a[2,0] = 0; a[2,2] = 1
            >>> a
            array([[[ 1.,  1.,  1.,  1.],
                    [ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.]],
            <BLANKLINE>
                   [[ 1.,  1.,  1.,  1.],
                    [ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.]],
            <BLANKLINE>
                   [[ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.],
                    [ 1.,  1.,  1.,  1.]],
            <BLANKLINE>
                   [[ 1.,  1.,  1.,  1.],
                    [ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.]],
            <BLANKLINE>
                   [[ 1.,  1.,  1.,  1.],
                    [ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.]]])

            >>> register_mean_offsets(a, include_shift=True)
            (masked_array(data =
             [[[1.0 1.0 1.0 1.0]
              [0.0 0.0 0.0 0.0]
              [0.0 0.0 0.0 0.0]]
            <BLANKLINE>
             [[1.0 1.0 1.0 1.0]
              [0.0 0.0 0.0 0.0]
              [0.0 0.0 0.0 0.0]]
            <BLANKLINE>
             [[-- -- -- --]
              [0.0 0.0 0.0 0.0]
              [0.0 0.0 0.0 0.0]]
            <BLANKLINE>
             [[1.0 1.0 1.0 1.0]
              [0.0 0.0 0.0 0.0]
              [0.0 0.0 0.0 0.0]]
            <BLANKLINE>
             [[1.0 1.0 1.0 1.0]
              [0.0 0.0 0.0 0.0]
              [0.0 0.0 0.0 0.0]]],
                         mask =
             [[[False False False False]
              [False False False False]
              [False False False False]]
            <BLANKLINE>
             [[False False False False]
              [False False False False]
              [False False False False]]
            <BLANKLINE>
             [[ True  True  True  True]
              [False False False False]
              [False False False False]]
            <BLANKLINE>
             [[False False False False]
              [False False False False]
              [False False False False]]
            <BLANKLINE>
             [[False False False False]
              [False False False False]
              [False False False False]]],
                   fill_value = 0.0)
            , array([[0, 0],
                   [0, 0],
                   [1, 0],
                   [0, 0],
                   [0, 0]]))
    """

    float_type = numpy.dtype(float_type).type

    # Must be of type float and must be at least 32-bit (smallest complex type
    # uses two 32-bit floats).
    assert issubclass(float_type, numpy.floating)
    assert numpy.dtype(float_type).itemsize >= 4

    # Sadly, there is no easier way to map the two types; so, this is it.
    float_complex_mapping = {
        numpy.float32 : numpy.complex64,
        numpy.float64 : numpy.complex128,
        numpy.float128 : numpy.complex256
    }
    complex_type = float_complex_mapping[float_type]

    if block_frame_length == -1:
        block_frame_length = len(frames2reg)

    tempdir_name = ""
    temporaries_filename = ""
    if isinstance(frames2reg, h5py.Dataset):
        tempdir_name, temporaries_filename = os.path.split(
            os.path.abspath(frames2reg.file.filename)
        )

        temporaries_filename = os.path.splitext(temporaries_filename)[0]
        temporaries_filename += "_".join(
            [
                frames2reg.name.replace("/", "_"),
                "temporaries.h5"
            ]
        )
        temporaries_filename = os.path.join(
            tempdir_name,
            temporaries_filename
        )
    elif (block_frame_length != len(frames2reg)):
        tempdir_name = tempfile.mkdtemp()
        temporaries_filename = os.path.join(tempdir_name, "temporaries.h5")

    frames2reg_fft = None
    space_shift = None
    this_space_shift = None
    if tempdir_name:
        temporaries_file = h5py.File(temporaries_filename, "w")

        frames2reg_fft = temporaries_file.create_dataset(
            "frames2reg_fft", shape=frames2reg.shape, dtype=complex_type
        )
        space_shift = temporaries_file.create_dataset(
            "space_shift",
            shape=(len(frames2reg), len(frames2reg.shape)-1),
            dtype=int
        )
        this_space_shift = temporaries_file.create_dataset(
            "this_space_shift",
            shape=space_shift.shape,
            dtype=space_shift.dtype
        )
    else:
        frames2reg_fft = numpy.empty(frames2reg.shape, dtype=complex_type)
        space_shift = numpy.zeros(
            (len(frames2reg), len(frames2reg.shape)-1), dtype=int
        )
        this_space_shift = numpy.empty_like(space_shift)

    for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
        frames2reg_fft[range_ij] = fft.fftn(
            frames2reg[range_ij], axes=range(1, len(frames2reg.shape))
        )

    template_fft = numpy.empty(frames2reg.shape[1:], dtype=complex_type)

    this_space_shift_mean = numpy.empty(
        this_space_shift.shape[1:],
        dtype=this_space_shift.dtype
    )

    # Repeat shift calculation until there is no further adjustment.
    num_iters = 0
    squared_magnitude_delta_space_shift = 1.0
    while (squared_magnitude_delta_space_shift != 0.0):
        squared_magnitude_delta_space_shift = 0.0

        template_fft[:] = 0
        for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
            template_fft += translate_fourier(
                frames2reg_fft[range_ij] / len(frames2reg),
                space_shift[range_ij]
            ).sum(axis=0)

        for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
            this_space_shift[range_ij] = find_offsets(
                frames2reg_fft[range_ij], template_fft
            )

        # Remove global shifts.
        this_space_shift_mean[...] = 0
        for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
            this_space_shift_mean += this_space_shift[range_ij].sum(axis=0)
        this_space_shift_mean[...] = numpy.round(
            this_space_shift_mean.astype(float_type) / len(this_space_shift)
        ).astype(this_space_shift_mean.dtype)
        for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
            this_space_shift[range_ij] = xnumpy.find_relative_offsets(
                this_space_shift[range_ij],
                this_space_shift_mean
            )

        # Find the shortest roll possible (i.e. if it is going over halfway
        # switch direction so it will go less than half).
        # Note all indices by definition were positive semi-definite and upper
        # bounded by the shape. This change will make them bound by
        # the half shape, but with either sign.
        for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
            this_space_shift[range_ij] = xnumpy.find_shortest_wraparound(
                this_space_shift[range_ij],
                frames2reg_fft.shape[1:]
            )

        for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
            delta_space_shift_ij = this_space_shift[range_ij] - \
                                   space_shift[range_ij]
            squared_magnitude_delta_space_shift += numpy.dot(
                delta_space_shift_ij, delta_space_shift_ij.T
            ).sum()

        for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
            space_shift[range_ij] = this_space_shift[range_ij]

        num_iters += 1
        logger.info(
            "Completed iteration, %i, " %
            num_iters
            + "where the L_2 norm squared of the relative shift was, %f." %
            squared_magnitude_delta_space_shift
        )
        if (max_iters != -1) and (num_iters >= max_iters):
            logger.info("Hit maximum number of iterations.")
            break

    reg_frames_shape = frames2reg.shape
    if to_truncate:
        space_shift_max = numpy.zeros(
            space_shift.shape[1:], dtype=space_shift.dtype
        )
        space_shift_min = numpy.zeros(
            space_shift.shape[1:], dtype=space_shift.dtype
        )
        for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
            numpy.maximum(
                space_shift_max,
                space_shift[range_ij].max(axis=0),
                out=space_shift_max
            )
            numpy.minimum(
                space_shift_min,
                space_shift[range_ij].min(axis=0),
                out=space_shift_min
            )
        reg_frames_shape = numpy.asarray(reg_frames_shape)
        reg_frames_shape[1:] -= space_shift_max
        reg_frames_shape[1:] += space_shift_min
        reg_frames_shape = tuple(reg_frames_shape)

        space_shift_max = tuple(space_shift_max)
        space_shift_min = space_shift_min.astype(object)
        space_shift_min[space_shift_min == 0] = None
        space_shift_min = tuple(space_shift_min)
        reg_frames_slice = tuple(
            slice(_1, _2) for _1, _2 in iters.izip(
                space_shift_max, space_shift_min
            )
        )

    # Adjust the registered frames using the translations found.
    # Mask rolled values.
    reg_frames = None
    if tempdir_name:
        if to_truncate:
            reg_frames = temporaries_file.create_dataset(
                "reg_frames",
                shape=reg_frames_shape,
                dtype=frames2reg.dtype,
                chunks=True
            )
        else:
            reg_frames = temporaries_file.create_group("reg_frames")
            reg_frames = hdf5.serializers.HDF5MaskedDataset(
                reg_frames, shape=frames2reg.shape, dtype=frames2reg.dtype
            )
    else:
        if to_truncate:
            reg_frames = numpy.empty(reg_frames_shape, dtype=frames2reg.dtype)
        else:
            reg_frames = numpy.ma.empty_like(frames2reg)
            reg_frames.mask = numpy.ma.getmaskarray(reg_frames)
            reg_frames.set_fill_value(reg_frames.dtype.type(0))

    for range_ij in iters.subrange(0, len(frames2reg), block_frame_length):
        for k in range_ij:
            if to_truncate:
                reg_frames[k] = xnumpy.roll(
                    frames2reg[k], space_shift[k]
                )[reg_frames_slice]
            else:
                reg_frames[k] = xnumpy.roll(
                    frames2reg[k], space_shift[k], to_mask=True
                )

    result = None
    results_filename = ""
    if tempdir_name:
        result = results_filename
        results_filename = os.path.join(tempdir_name, "results.h5")
        results_file = h5py.File(results_filename, "w")
        if to_truncate:
            temporaries_file.copy(reg_frames.name, results_file)
        else:
            temporaries_file.copy(reg_frames.group, results_file)
        if include_shift:
            temporaries_file.copy(space_shift.name, results_file)
        frames2reg_fft = None
        reg_frames = None
        space_shift = None
        this_space_shift = None
        temporaries_file.close()
        os.remove(temporaries_filename)
        temporaries_filename = ""
        result = results_filename
    else:
        result = reg_frames
        if include_shift:
            result = (reg_frames, space_shift)

    if tempdir_name:
        results_file.close()
        results_file = None

    return(result)


@prof.log_call(trace_logger)
@lru_cache(maxsize=2)
def generate_unit_phase_shifts(shape, float_type=float):
    """
        Computes the complex phase shift's angle due to a unit spatial shift.

        This is meant to be a helper function for ``register_mean_offsets``. It
        does this by computing a table of the angle of the phase of a unit
        shift in each dimension (with a factor of :math:`2\pi`).

        This allows arbitrary phase shifts to be made in each dimensions by
        multiplying these angles by the size of the shift and added to the
        existing angle to induce the proper phase shift in fourier space, which
        is equivalent to the spatial translation.

        Args:
            shape(tuple of ints):       shape of the data to be shifted.

            float_type(real type):      phase type (default numpy.float64)

        Returns:
            (numpy.ndarray):            an array containing the angle of the
                                        complex phase shift to use for each
                                        dimension.

        Examples:
            >>> generate_unit_phase_shifts((2,4))
            array([[[-0.        , -0.        , -0.        , -0.        ],
                    [-3.14159265, -3.14159265, -3.14159265, -3.14159265]],
            <BLANKLINE>
                   [[-0.        , -1.57079633, -3.14159265, -4.71238898],
                    [-0.        , -1.57079633, -3.14159265, -4.71238898]]])
    """

    # Convert to `numpy`-based type if not done already.
    float_type = numpy.dtype(float_type).type

    # Must be of type float.
    assert issubclass(float_type, numpy.floating)
    assert numpy.dtype(float_type).itemsize >= 4

    # Get the negative wave vector
    negative_wave_vector = numpy.asarray(shape, dtype=float_type)
    numpy.reciprocal(negative_wave_vector, out=negative_wave_vector)
    negative_wave_vector *= 2*numpy.pi
    numpy.negative(negative_wave_vector, out=negative_wave_vector)

    # Get the indices for each point in the selected space.
    indices = xnumpy.cartesian_product([numpy.arange(_) for _ in shape])

    # Determine the phase offset for each point in space.
    complex_angle_unit_shift = indices * negative_wave_vector
    complex_angle_unit_shift = complex_angle_unit_shift.T.copy()
    complex_angle_unit_shift = complex_angle_unit_shift.reshape(
        (len(shape),) + shape
    )

    return(complex_angle_unit_shift)


@prof.log_call(trace_logger)
def translate_fourier(frame_fft, shift):
    """
        Translates frame(s) of data in Fourier space using the shift(s) given.

        Args:
            frame_fft(complex array):   Either a single frame with C-order axes
                                        or multiple frames with time on the 0th
                                        axis.

            shift(array of ints):       Either the shift for each dimension
                                        with C-ordered values or multiple
                                        frames with time on the 0th axis.

        Returns:
            (numpy.ndarray):            The frame(s) shifted.

        Examples:
            >>> a = numpy.arange(12).reshape(3,4).astype(float)
            >>> a
            array([[  0.,   1.,   2.,   3.],
                   [  4.,   5.,   6.,   7.],
                   [  8.,   9.,  10.,  11.]])
            >>> af = fft.fftn(a, axes=tuple(iters.irange(a.ndim)))
            >>> numpy.around(af, decimals=10)
            array([[ 66. +0.j        ,  -6. +6.j        ,  -6. +0.j        ,  -6. -6.j        ],
                   [-24.+13.85640646j,   0. +0.j        ,   0. +0.j        ,   0. +0.j        ],
                   [-24.-13.85640646j,   0. +0.j        ,   0. +0.j        ,   0. +0.j        ]])

            >>> s = numpy. array([1, -1])

            >>> atf = translate_fourier(af, s)
            >>> numpy.around(atf, decimals=10)
            array([[ 66. +0.j        ,  -6. -6.j        ,   6. -0.j        ,  -6. +6.j        ],
                   [ 24.+13.85640646j,   0. +0.j        ,   0. +0.j        ,  -0. +0.j        ],
                   [ 24.-13.85640646j,   0. -0.j        ,   0. +0.j        ,   0. +0.j        ]])

            >>> fft.ifftn(
            ...     atf, axes=tuple(iters.irange(a.ndim))
            ... ).real.round().astype(int).astype(float)
            array([[  9.,  10.,  11.,   8.],
                   [  1.,   2.,   3.,   0.],
                   [  5.,   6.,   7.,   4.]])

            >>> a = a[None]; af = af[None]; s = s[None]
            >>> atf = translate_fourier(af, s)
            >>> numpy.around(atf, decimals=10)
            array([[[ 66. +0.j        ,  -6. -6.j        ,   6. -0.j        ,  -6. +6.j        ],
                    [ 24.+13.85640646j,   0. +0.j        ,   0. +0.j        ,  -0. +0.j        ],
                    [ 24.-13.85640646j,   0. -0.j        ,   0. +0.j        ,   0. +0.j        ]]])


            >>> fft.ifftn(
            ...     atf, axes=tuple(iters.irange(1, a.ndim))
            ... ).real.round().astype(int).astype(float)
            array([[[  9.,  10.,  11.,   8.],
                    [  1.,   2.,   3.,   0.],
                    [  5.,   6.,   7.,   4.]]])

    """

    add_frame_axis = False
    if (len(shift.shape) == 1) and (len(shift) == len(frame_fft.shape)):
        add_frame_axis = True
        shift = shift[None]
        frame_fft = frame_fft[None]

    assert (
        (len(shift.shape) == 2) and
        (shift.shape[1] == (len(frame_fft.shape) - 1))
    ), "Shapes are incompatible." + \
       ("`shift.shape = %s`" % repr(shift.shape)) + \
       (" and `frame_fft.shape = %s`." % repr(frame_fft.shape))

    # Sadly, there is no easier way to map the two types; so, this is it.
    complex_type = frame_fft.dtype.type
    complex_float_mapping = {
        numpy.complex64 : numpy.float32,
        numpy.complex128 : numpy.float64,
        numpy.complex256 : numpy.float128
    }
    float_type = complex_float_mapping[complex_type]
    J = complex_type(1j)

    # Get unit translations in all directions as the complex phase's angle.
    unit_space_shift_fft = generate_unit_phase_shifts(
        frame_fft.shape[1:], float_type=float_type
    )

    # Compute phase adjustment in complex.
    frame_fft_shifted = numpy.exp(
        J * numpy.tensordot(
                shift,
                unit_space_shift_fft,
                axes=[-1, 0]
            )
    )
    frame_fft_shifted *= frame_fft

    if add_frame_axis:
        frame_fft_shifted = frame_fft_shifted[0]

    return(frame_fft_shifted)


@prof.log_call(trace_logger)
def find_offsets(frames2reg_fft, template_fft):
    """
        Computes the convolution of the template with the frames by taking
        advantage of their FFTs for faster computation that an ordinary
        convolution ( O(N*lg(N)) vs O(N^2) )
        < http://ccrma.stanford.edu/~jos/ReviewFourier/FFT_Convolution_vs_Direct.html >.
        Once computed the maximum of the convolution is found to determine the
        best overlap of each frame with the template, which provides the needed
        offset. Some corrections are performed to make reasonable offsets.

        Notes:
            Adapted from code provided by Wenzhi Sun with speed improvements
            provided by Uri Dubin.

        Args:
            frames2reg(numpy.ndarray):           image stack to register (time
                                                 is the first dimension uses
                                                 C-order tyx or tzyx).

            template_fft(numpy.ndarray):         what to register the image
                                                 stack against (single frame
                                                 using C-order yx or zyx).

        Returns:
            (numpy.ndarray):                     an array containing the
                                                 translations to apply to each
                                                 frame.

        Examples:
            >>> a = numpy.zeros((5, 3, 4)); a[:,0] = 1; a[2,0] = 0; a[2,2] = 1
            >>> a
            array([[[ 1.,  1.,  1.,  1.],
                    [ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.]],
            <BLANKLINE>
                   [[ 1.,  1.,  1.,  1.],
                    [ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.]],
            <BLANKLINE>
                   [[ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.],
                    [ 1.,  1.,  1.,  1.]],
            <BLANKLINE>
                   [[ 1.,  1.,  1.,  1.],
                    [ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.]],
            <BLANKLINE>
                   [[ 1.,  1.,  1.,  1.],
                    [ 0.,  0.,  0.,  0.],
                    [ 0.,  0.,  0.,  0.]]])

            >>> af = numpy.fft.fftn(a, axes=range(1, a.ndim)); af
            array([[[ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ]],
            <BLANKLINE>
                   [[ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ]],
            <BLANKLINE>
                   [[ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [-2.+3.46410162j,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [-2.-3.46410162j,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ]],
            <BLANKLINE>
                   [[ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ]],
            <BLANKLINE>
                   [[ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ],
                    [ 4.+0.j        ,  0.+0.j        ,  0.+0.j        ,  0.+0.j        ]]])

            >>> tf = numpy.fft.fftn(a.mean(axis=0)); tf
            array([[ 4.0+0.j        ,  0.0+0.j        ,  0.0+0.j        ,  0.0+0.j        ],
                   [ 2.8+0.69282032j,  0.0+0.j        ,  0.0+0.j        ,  0.0+0.j        ],
                   [ 2.8-0.69282032j,  0.0+0.j        ,  0.0+0.j        ,  0.0+0.j        ]])

            >>> find_offsets(af, tf)
            array([[ 0,  0],
                   [ 0,  0],
                   [-2,  0],
                   [ 0,  0],
                   [ 0,  0]])
    """

    # If there is only one frame, add a singleton axis to indicate this.
    frames2reg_fft_added_singleton = (frames2reg_fft.ndim == template_fft.ndim)
    if frames2reg_fft_added_singleton:
        frames2reg_fft = frames2reg_fft[None]

    # Compute the product of the two FFTs (i.e. the convolution of the regular
    # versions).
    frames2reg_template_conv_fft = frames2reg_fft * template_fft.conj()[None]

    # Find the FFT inverse (over all spatial dimensions) to return to the
    # convolution.
    frames2reg_template_conv = fft.ifftn(
        frames2reg_template_conv_fft, axes=range(1, frames2reg_fft.ndim)
    )

    # Find where the convolution is maximal. Will have the most things in
    # common between the template and frames.
    frames2reg_template_conv_max, frames2reg_template_conv_max_indices = xnumpy.max_abs(
        frames2reg_template_conv,
        axis=range(1, frames2reg_fft.ndim),
        return_indices=True
    )

    # First index is just the frame, which will be in sequential order. We
    # don't need this so we drop it.
    frames2reg_template_conv_max_indices = frames2reg_template_conv_max_indices[1:]

    # Convert indices into an array for easy manipulation.
    frames2reg_template_conv_max_indices = numpy.array(
        frames2reg_template_conv_max_indices
    ).T.copy()

    # Shift will have to be in the opposite direction to bring everything to
    # the center.
    numpy.negative(
        frames2reg_template_conv_max_indices,
        out=frames2reg_template_conv_max_indices
    )

    return(frames2reg_template_conv_max_indices)



"""
The ``spams_sandbox`` module provides mechanisms for sandboxing SPAMS.

===============================================================================
Overview
===============================================================================
SPAMS sometimes seems to step on the interpreter. As a result, we provide a
number of strategies to address this, by launching it in a separate process so
that it hopefully does not mess up the main interpreter. We also try to keep
the module space clean. This seems to help.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Jun 20, 2014 12:07:48 EDT$"



class SPAMSException(Exception):
    pass


def run_multiprocessing_queue_spams_trainDL(out_queue, *args, **kwargs):
    """
        Designed to run spams.trainDL in a separate process.

        It is necessary to run SPAMS in a separate process as segmentation
        faults have been discovered in later parts of the Python code dependent
        on whether SPAMS has run or not. It is suspected that spams may
        interfere with the interpreter. Thus, it should be sandboxed (run in a
        different Python interpreter) so that it doesn't damage what happens in
        this one.

        This particular version uses a multiprocessing.Queue to return the
        resulting dictionary.


        Args:
            out_queue(multiprocessing.Queue):       what will take the returned
                                                    dictionary from
                                                    spams.trainDL.

            *args(list):                            a list of position
                                                    arguments to pass to
                                                    spams.trainDL.

            **kwargs(dict):                         a dictionary of keyword
                                                    arguments to pass to
                                                    spams.trainDL.
    """

    # It is not needed outside of calling spams.trainDL.
    # Also, it takes a long time to load this module.
    import spams

    result = spams.trainDL(*args, **kwargs)
    out_queue.put(result)


def call_multiprocessing_queue_spams_trainDL(*args, **kwargs):
    """
        Designed to start spams.trainDL in a separate process and handle the
        result in an unnoticeably different way.

        It is necessary to run SPAMS in a separate process as segmentation
        faults have been discovered in later parts of the Python code dependent
        on whether SPAMS has run or not. It is suspected that spams may
        interfere with the interpreter. Thus, it should be sandboxed (run in a
        different Python interpreter) so that it doesn't damage what happens in
        this one.

        This particular version uses a multiprocessing.Queue to return the
        resulting dictionary.


        Args:
            *args(list):                            a list of position
                                                    arguments to pass to
                                                    spams.trainDL.

            **kwargs(dict):                         a dictionary of keyword
                                                    arguments to pass to
                                                    spams.trainDL.

        Returns:
            result(numpy.matrix): the dictionary found
    """

    # Only necessary for dealing with SPAMS
    import multiprocessing

    out_queue = multiprocessing.Queue()

    queue_args = (out_queue,) + args
    p = multiprocessing.Process(
        target=run_multiprocessing_queue_spams_trainDL,
        args=queue_args,
        kwargs=kwargs
    )
    p.start()
    result = out_queue.get()
    result = result.copy()
    p.join()

    if p.exitcode != 0: raise SPAMSException(
        "SPAMS has terminated with exitcode \"" + repr(p.exitcode) + "\"."
    )

    return(result)


def run_multiprocessing_array_spams_trainDL(result_array_type,
                                            result_array,
                                            X_array_type,
                                            X_array,
                                            D_is_arg=False,
                                            D_array_type=None,
                                            D_array=None,
                                            *args,
                                            **kwargs):
    """
        Designed to start spams.trainDL in a separate process and handle the
        result in an unnoticeably different way.

        It is necessary to run SPAMS in a separate process as segmentation
        faults have been discovered in later parts of the Python code dependent
        on whether SPAMS has run or not. It is suspected that spams may
        interfere with the interpreter. Thus, it should be sandboxed (run in a
        different Python interpreter) so that it doesn't damage what happens in
        this one.

        This particular version uses a multiprocessing.Array to share memory to
        return the resulting dictionary.


        Args:
            result_array_type(numpy.ctypeslib.ndpointer):   a pointer type with
                                                            properties needed
                                                            by result_array.

            result_array(multiprocessing.RawArray):         shared memory array
                                                            to store results
                                                            in.

            X_array_type(numpy.ctypeslib.ndpointer):        a pointer type with
                                                            properties needed
                                                            by X_array.

            X_array(numpy.ndarray):                         currently uses
                                                            numpy ndarray as
                                                            input.

            *args(list):                                    a list of position
                                                            arguments to pass
                                                            to spams.trainDL.

            **kwargs(dict):                                 a dictionary of
                                                            keyword arguments
                                                            to pass to
                                                            spams.trainDL.

        Note:
            This is somewhat faster than using multiprocessing.Queue.
    """

    # Just to make sure this exists in the new process. Shouldn't be necessary.
    import numpy
    # Just to make sure this exists in the new process. Shouldn't be necessary.
    # Also, it is not needed outside of calling this function.
    import spams

    as_ordered_array_dict = {
        "F_CONTIGUOUS": numpy.asfortranarray,
        "C_CONTIGUOUS": numpy.ascontiguousarray
    }

    # Construct X from shared array.
    X_dtype = X_array_type._dtype_
    X_shape = X_array_type._shape_
    X_flags = numpy.core.multiarray.flagsobj(X_array_type._flags_)

    X = numpy.frombuffer(X_array, dtype=X_dtype).reshape(X_shape)
    X.setflags(X_flags)

    for order_name, as_ordered_array in as_ordered_array_dict.items():
        if order_name in X_array_type.__name__:
            X = as_ordered_array(X)

    # Construct D from shared array.
    if (D_array_type is not None) and (D_array is not None):
        D_dtype = D_array_type._dtype_
        D_shape = D_array_type._shape_
        D_flags = numpy.core.multiarray.flagsobj(D_array_type._flags_)

        D = numpy.frombuffer(D_array, dtype=D_dtype).reshape(D_shape)
        D.setflags(D_flags)

        for order_name, as_ordered_array in as_ordered_array_dict.items():
            if order_name in D_array_type.__name__:
                D = as_ordered_array(D)

        if D_is_arg:
            args[3] = D
        else:
            kwargs["D"] = D

    # Construct the result to use the shared buffer.
    result_dtype = result_array_type._dtype_
    result_shape = result_array_type._shape_
    result_flags = numpy.core.multiarray.flagsobj(result_array_type._flags_)

    result = numpy.frombuffer(
        result_array, dtype=result_dtype
    ).reshape(result_shape)
    result.setflags(result_flags)

    for order_name, as_ordered_array in as_ordered_array_dict.items():
        if order_name in result_array_type.__name__:
            result = as_ordered_array(result)

    result[:] = spams.trainDL(X, *args, **kwargs)


def call_multiprocessing_array_spams_trainDL(X, *args, **kwargs):
    """
        Designed to start spams.trainDL in a separate process and handle
        result in an unnoticeably different way.

        It is necessary to run SPAMS in a separate process as segmentation
        faults have been discovered in later parts of the Python code dependent
        on whether SPAMS has run or not. It is suspected that spams may
        interfere with the interpreter. Thus, it should be sandboxed (run in a
        different Python interpreter) so that it doesn't damage what happens in
        this one.

        This particular version uses a multiprocessing.Array to share memory to
        return the resulting dictionary.


        Args:
            X(numpy.matrix):                        a Fortran order NumPy
                                                    Matrix with the same name
                                                    as used by spams.trainDL
                                                    (so if someone tries to use
                                                    it as a keyword
                                                    argument...).

            *args(list):                            a list of position
                                                    arguments to pass to
                                                    spams.trainDL.

            **kwargs(dict):                         a dictionary of keyword
                                                    arguments to pass to
                                                    spams.trainDL.

        Note:
            This is somewhat faster than using multiprocessing.Queue.
    """

    # Only necessary for dealing with SPAMS
    import multiprocessing

    # Just to make sure this exists in the new process. Shouldn't be necessary.
    import numpy

    D_is_arg = False
    D = None
    if (len(args) >= 4):
        D_is_arg = True
        D = args[3]
        args[3] = None
    else:
        D = kwargs.pop("D", None)

    # Types for X_array
    X_array_type = numpy.ctypeslib.ndpointer(
        dtype=X.dtype, ndim=X.ndim, shape=X.shape, flags=X.flags
    )
    X_array_ctype = type(
        numpy.ctypeslib.as_ctypes(X_array_type._dtype_.type(0)[()])
    )

    # Create a shared array to contain X
    X_array = multiprocessing.Array(X_array_ctype,
                                    X.size,
                                    lock=False)

    # Copy over the contents of X.
    X_array_numpy = numpy.frombuffer(
        X_array, dtype=X_array_type._dtype_
    ).reshape(X_array_type._shape_)
    X_array_numpy[:] = X
    X_array_numpy = None

    len_D = kwargs.get("K", None)
    if D is not None:
        # Types for D_array
        D_array_type = numpy.ctypeslib.ndpointer(
            dtype=D.dtype, ndim=D.ndim, shape=D.shape, flags=D.flags
        )
        D_array_ctype = type(
            numpy.ctypeslib.as_ctypes(D_array_type._dtype_.type(0)[()])
        )

        # Create a shared array to contain D
        D_array = multiprocessing.Array(D_array_ctype,
                                        D.size,
                                        lock=False)

        # Copy over the contents of D.
        D_array_numpy = numpy.frombuffer(
            D_array, dtype=D_array_type._dtype_
        ).reshape(D_array_type._shape_)
        D_array_numpy[:] = D
        D_array_numpy = None

        len_D = D.shape[-1]

    # Types for result_array
    result_array_type = numpy.ctypeslib.ndpointer(
        dtype=X.dtype, ndim=X.ndim, shape=(X.shape[0], len_D)
    )
    result_array_ctype = type(
        numpy.ctypeslib.as_ctypes(result_array_type._dtype_.type(0)[()])
    )

    # Create a shared array to contain the result
    result_array = multiprocessing.Array(
        result_array_ctype,
        int(numpy.product(result_array_type._shape_)),
        lock=False
    )

    new_args = (
        result_array_type, result_array,
        X_array_type, X_array,
    )
    if D is not None:
        new_args = new_args + (
            D_is_arg, D_array_type, D_array,
        )
    p = multiprocessing.Process(
        target=run_multiprocessing_array_spams_trainDL,
        args=new_args,
        kwargs=kwargs
    )
    p.start()
    p.join()

    if p.exitcode != 0: raise SPAMSException(
        "SPAMS has terminated with exitcode \"" + repr(p.exitcode) + "\"."
    )

    # Reconstruct the result from the output array
    result = numpy.frombuffer(
        result_array, dtype=result_array_type._dtype_
    ).reshape(result_array_type._shape_)
    result = result.copy()

    return(result)


def call_spams_trainDL(*args, **kwargs):
    """
        Encapsulates call to spams.trainDL. Ensures copy of results occur just
        in case. Designed to be like the multiprocessing calls.

        Args:
            *args(list):                            a list of position
                                                    arguments to pass to
                                                    spams.trainDL.

            **kwargs(dict):                         a dictionary of keyword
                                                    arguments to pass to
                                                    spams.trainDL.

        Note:
            For legacy.
    """

    # It is not needed outside of calling spams.trainDL.
    # Also, it takes a long time to load this module.
    import spams

    result = spams.trainDL(*args, **kwargs)
    result = result.copy()

    return(result)
